"""Toml Sort

Provides a class to simply sort TOMl files
"""

import re
import itertools
from typing import Iterable, Tuple, Set

import tomlkit
from tomlkit.api import aot, ws, item
from tomlkit.toml_document import TOMLDocument
from tomlkit.container import Container
from tomlkit.items import Item, Table, AoT, Comment, Whitespace, Trivia


def clean_toml_text(input_toml: str) -> str:
    """Clean input toml, increasing the chance for beautiful output"""
    cleaned = re.sub(r"[\r\n][\r\n]{2,}", "\n\n", input_toml)
    return "\n" + cleaned.strip() + "\n"


def write_header_comment(from_doc: TOMLDocument, to_doc: TOMLDocument) -> None:
    """Write header comment from the FROM doc to the TO doc

    Only writes comments / whitespace from the beginning of a TOML document.
    Anything from later in the document is ambiguous and cannot be sorted
    accurately.
    """
    for _, value in from_doc.body:
        if isinstance(value, Whitespace):
            to_doc.add(ws("\n"))
        elif isinstance(value, Comment):
            value.trivia.indent = ""
            to_doc.add(Comment(value.trivia))
        else:
            to_doc.add(ws("\n"))
            return


def convert_tomlkit_buggy_types(in_value: object, parent: Item) -> Item:
    """Fix buggy items while iterating through tomlkit items

    Iterating through tomlkit items can often be buggy; this function fixes it
    """
    if isinstance(in_value, Item):
        return in_value
    if isinstance(in_value, Container):
        return Table(in_value, trivia=Trivia(), is_aot_element=False)
    return item(in_value, parent)


class TomlSort:
    """API to manage sorting toml files"""

    def __init__(
        self,
        input_toml: str,
        only_sort_tables: bool = False,
        no_header: bool = False,
    ) -> None:
        """Initializer

        :attr input_toml: the input toml data for processing
        :attr only_sort_tables: turns on sorting for only tables
        :attr no_header: omit leading comments from the output
        """
        self.input_toml = input_toml
        self.only_sort_tables = only_sort_tables
        self.no_header = no_header

    def sorted_children_table(
        self, parent: Table
    ) -> Iterable[Tuple[str, Item]]:
        """Get the sorted children of a table

        NOTE: non-tables are wrapped in an item to ensure that they are, in
        fact, items. Tables and AoT's are definitely items, so the conversion
        is not necessary.
        """
        table_items = [
            (key, convert_tomlkit_buggy_types(parent[key], parent))
            for key in parent.keys()
        ]
        tables = (
            (key, value)
            for key, value in table_items
            if isinstance(value, (Table, AoT))
        )
        non_tables = (
            (key, value)
            for key, value in table_items
            if not isinstance(value, (Table, AoT))
        )
        non_tables_final = (
            sorted(non_tables, key=lambda x: x[0])
            if not self.only_sort_tables
            else non_tables
        )
        return itertools.chain(
            non_tables_final, sorted(tables, key=lambda x: x[0])
        )

    def toml_elements_sorted(self, original: Item) -> Item:
        """Returns a sorted item, recursing collections to their base"""
        if isinstance(original, Table):
            original.trivia.indent = "\n"
            new_table = Table(
                Container(),
                trivia=original.trivia,
                is_aot_element=original.is_aot_element(),
                is_super_table=original.is_super_table(),
            )
            for key, value in self.sorted_children_table(original):
                new_table[key] = self.toml_elements_sorted(value)
            return new_table
        if isinstance(original, AoT):
            # REFACTOR WHEN POSSIBLE:
            #
            # NOTE: this section contains a necessary hack. Tomlkit's AoT
            # implementation currently generates duplicate elements. I rely on
            # the object id to remove these duplicates; the object id will
            # allow correct duplicates to remain
            new_aot = aot()
            id_lookup = set()  # type: Set[int]
            for aot_item in original:
                id_aot_item = id(aot_item)
                if id_aot_item not in id_lookup:
                    new_aot.append(self.toml_elements_sorted(aot_item))
                    id_lookup.add(id_aot_item)
            return new_aot
        if isinstance(original, Item):
            original.trivia.indent = ""
            return original
        raise TypeError(
            "Invalid TOML; " + str(type(original)) + " is not an Item."
        )

    def toml_doc_sorted(self, original: TOMLDocument) -> TOMLDocument:
        """Sort a TOMLDocument"""
        sorted_document = tomlkit.document()
        if not self.no_header:
            write_header_comment(original, sorted_document)
        for key, value in self.sorted_children_table(original):
            sorted_document[key] = self.toml_elements_sorted(value)
        return sorted_document

    def sorted(self) -> str:
        """Sort a TOML string"""
        clean_toml = clean_toml_text(self.input_toml)
        toml_doc = tomlkit.parse(clean_toml)
        sorted_toml = self.toml_doc_sorted(toml_doc)
        return clean_toml_text(tomlkit.dumps(sorted_toml)).strip() + "\n"



"""Toml Sort CLI"""

import sys
import click

from . import TomlSort


@click.command()
@click.option(
    "-o",
    "--output",
    type=click.File("w"),
    default="-",
    show_default=True,
    help="The output filepath. Choose stdout with '-'.",
)
@click.option(
    "-a",
    "--all",
    "_all",
    is_flag=True,
    help=(
        "Sort all keys. "
        "Default is to only sort non-inline 'tables and arrays of tables'."
    ),
)
@click.option(
    "--no-header",
    is_flag=True,
    help="Do not keep a document's leading comments.",
)
@click.argument("filename", type=click.File("r"), default="-")
@click.version_option()
def cli(output, _all, no_header, filename) -> None:
    """Sort toml file FILENAME, saving results to a file, or stdout (default)

    FILENAME a filepath or standard input (-)

    Examples (non-exhaustive list):

        Read from stdin, write to stdout: cat input.toml | toml-sort

        Read from disk, write to disk   : toml-sort -o output.toml input.toml

        Read from disk, write to stdout : toml-sort input.toml
    """
    if filename.isatty():
        error_message_if_terminal = """
toml-sort: missing FILENAME, and no stdin
Usage: toml-sort [OPTIONS] [FILENAME]

Try `toml-sort --help` for more information
""".strip()
        click.echo(error_message_if_terminal, err=True)
        sys.exit(1)
    sorted_toml = TomlSort(
        input_toml=filename.read(),
        only_sort_tables=not bool(_all),
        no_header=bool(no_header),
    ).sorted()
    output.write(sorted_toml)



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['toml_sort']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.0,<8.0', 'tomlkit>=0.5.5']

entry_points = \
{'console_scripts': ['toml-sort = toml_sort.cli:cli']}

setup_kwargs = {
    'name': 'toml-sort',
    'version': '0.15.1',
    'description': 'Toml sorting library',
    'long_description': '# toml-sort\n\n[![image-version](https://img.shields.io/pypi/v/toml-sort.svg)](https://python.org/pypi/toml-sort)\n[![image-license](https://img.shields.io/pypi/l/toml-sort.svg)](https://python.org/pypi/toml-sort)\n[![image](https://img.shields.io/pypi/pyversions/toml-sort.svg)](https://python.org/pypi/toml-sort)\n\nA command line utility to sort and format your toml files. Requires Python 3.6 or greater.\n\n## Installation\n\n```bash\n# With pip\npip install toml-sort\n\n# With poetry\npoetry add toml-sort\n```\n\n## Motivation\n\n This library strives to sort TOML files by providing the following features:\n\n* Sort tables and Arrays of Tables (AoT)\n* Option to sort non-tables / non-AoT\'s, or not\n* Preserve inline comments\n* Option to preserve top-level document comments, or not\n* Standardize whitespace and indentation\n\nI wrote this library because I couldn\'t find any "good" sorting utilities for TOML files.\n\n## Usage\n\nThis project can be used as either a command line utility or a Python library.\n\n### Command line interface\n\nPrint detailed help\n\n    toml-sort --help\n\nRead from stdin, write to stdout:\n\n    cat input.toml | toml-sort\n\nRead from file on disk, write to file on disk:\n\n    toml-sort -o output.toml input.toml\n\nRead from file on disk, write to stdout\n\n    toml-sort input.toml\n\nRead from stdin, write to file on disk\n\n    cat input.toml | toml-sort -o output.toml\n\n## Example\n\nThe following example shows the input, and output, from the CLI with default options.\n\n### Unformatted, unsorted input\n\n```toml\n# My great TOML example\n\n  title = "The example"\n\n[[a-section.hello]]\nports = [ 8001, 8001, 8002 ]\ndob = 1979-05-27T07:32:00Z # First class dates? Why not?\n\n\n\n  [b-section]\n  date = "2018"\n  name = "Richard Stallman"\n\n[[a-section.hello]]\nports = [ 80 ]\ndob = 1920-05-27T07:32:00Z # Another date!\n\n                          [a-section]\n                          date = "2019"\n                          name = "Samuel Roeca"\n```\n\n### Formatted, sorted output\n\n```toml\n# My great TOML example\n\ntitle = "The example"\n\n[a-section]\ndate = "2019"\nname = "Samuel Roeca"\n\n[[a-section.hello]]\nports = [ 8001, 8001, 8002 ]\ndob = 1979-05-27T07:32:00Z # First class dates? Why not?\n\n[[a-section.hello]]\nports = [ 80 ]\ndob = 1920-05-27T07:32:00Z # Another date!\n\n[b-section]\ndate = "2018"\nname = "Richard Stallman"\n```\n\n## Local Development\n\nLocal development for this project is quite simple.\n\n**Dependencies**\n\nInstall the following tools manually.\n\n* [Poetry](https://github.com/sdispater/poetry#installation)\n* [GNU Make](https://www.gnu.org/software/make/)\n\n*Recommended*\n\n* [pyenv](https://github.com/pyenv/pyenv)\n\n**Set up development environment**\n\n```bash\nmake setup\n```\n\n**Run Tests**\n\n```bash\nmake test\n```\n\n## Written by\n\nSamuel Roeca *samuel.roeca@gmail.com*\n',
    'author': 'Sam Roeca',
    'author_email': 'samuel.roeca@gmail.com',
    'url': 'https://github.com/pappasam/toml-sort',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
